"""Slices AI Infrastructure Service Job Debug Logs API."""

from uuid import UUID

from httpx import Client, HTTPError
from msgspec.json import Decoder as JsonDecoder
from slices_clientlib_core.base.exceptions import from_httpx_exception

from .schema.job import Job
from .schema.job_debug_logs import JobDebugLog

_job_debug_logs_decoder = JsonDecoder(list[JobDebugLog])


class AIJobDebugLogsClient:
    """AI Infrastructure Service Job API Debug Logs Client."""

    API_BASE_PATH = "/apis/ai.slices-ri.eu/v1/"

    def __init__(self, client: Client) -> None:
        """Create AI Infrastructure Service Jobs Client."""
        self._client = client

        self._base_path = f"{self.API_BASE_PATH}jobs/"

        assert client.base_url is not None, "Expected base_url to be set"

    def get(self, job: str | UUID | Job) -> list[JobDebugLog]:
        """Get job Debug Log."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        response = self._client.get(f"{self._base_path}{job_id}/debug/")
        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err
        else:
            return _job_debug_logs_decoder.decode(response.content)



"""Slices AI Infrastructure Service Resource API."""

from httpx import Client, HTTPError
from msgspec.json import Decoder as JsonDecoder
from slices_clientlib_core.base.exceptions import from_httpx_exception

from .schema.worker import Cluster, Slave

_slave_list_decoder = JsonDecoder(list[Slave])
_cluster_list_decoder = JsonDecoder(list[Cluster])


class AIWorkersClient:
    """AI Infrastructure Service Workers API Client."""

    API_BASE_PATH = "/apis/ai.slices-ri.eu/v1/"

    def __init__(self, client: Client) -> None:
        """Create AI Infrastructure Service Workers Client."""

        self._client = client

        self._base_path = f"{self.API_BASE_PATH}"

        assert client.base_url is not None, "Expected base_url to be set"

    def list_clusters(self) -> list[Cluster]:
        """List clusters."""

        response = self._client.get(f"{self._base_path}clusters")

        try:
            response.raise_for_status()
            return _cluster_list_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def list_slaves(self) -> list[Slave]:
        """List workers."""

        response = self._client.get(f"{self._base_path}slaves")

        try:
            response.raise_for_status()
            return _slave_list_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err



"""Schema for NewJob, used to create new jobs."""

from msgspec import Struct, field

from slices_clientlib_aiis.schema.job import (
    DurationStr,
    JobRequestBase,
    JobRequestExtra,
    JobRequestScheduling,
    JobState,
    RestartInfo,
)
from slices_clientlib_aiis.util.msgspec import NonEmptyStr


class NewJobRequestScheduling(JobRequestScheduling):
    """GPULab scheduling instructions for new Jobs.

    This has extra defaults, and allowes DurationStr in not_after and not_before.
    """

    min_duration: DurationStr = "2 hour"
    max_duration: DurationStr = "8 hour"
    # This is what we want
    # not_before: AwareDatetime | DurationStr | None = None
    # not_after: AwareDatetime | DurationStr | None = "1 week"
    # But msgspec does not support unions with multiple str types, so we use plain str
    not_before: str | None = None  # pyright: ignore [reportIncompatibleVariableOverride] should be OK for this usage
    not_after: str | None = "1 week"  # pyright: ignore [reportIncompatibleVariableOverride] should be OK for this usage


class NewJobRequest(JobRequestBase):
    """Job Request for new jobs (extra defaults)."""

    scheduling: NewJobRequestScheduling = field(default_factory=lambda: NewJobRequestScheduling())
    extra: JobRequestExtra = field(default_factory=lambda: JobRequestExtra())


class NewJob(Struct, rename="camel"):
    """New job: schema limited to the values that must be present during initial submission."""

    name: NonEmptyStr
    request: NewJobRequest
    description: NonEmptyStr | None = None
    state: JobState | None = None
    restart_info: RestartInfo | None = None


class NewJobReq(Struct):
    """Request schema for new job creation."""

    job: NewJob
    idtoken: str



"""Utilities related to MsgSpec."""

from datetime import datetime
from typing import Annotated

from msgspec import Meta
from msgspec.json import Encoder

json_encoder = Encoder()

# Useful msgspec types

AwareDatetime = Annotated[datetime, Meta(tz=True)]

PositiveInt = Annotated[int, Meta(gt=0)]
NonEmptyStr = Annotated[str, Meta(min_length=1)]

PublicSshKey = Annotated[str, Meta(min_length=20)]  # TODO
EmailStr = Annotated[str, Meta(min_length=5)]  # TODO



"""Slices AI Infrastructure Service Job API."""

from collections.abc import Collection
from datetime import datetime, timedelta
from enum import Enum
from functools import cached_property
from typing import Any
from uuid import UUID

from httpx import Client, HTTPError
from msgspec.json import Decoder as JsonDecoder
from slices_clientlib_core.base.config import ApiEntry
from slices_clientlib_core.base.exceptions import from_httpx_exception
from slices_clientlib_core.client import (
    SlicesClient as SlicesCoreClient,
)
from slices_clientlib_core.client import (
    SlicesExperimentClient as SlicesCoreExperimentClient,
)
from slices_clientlib_core.client import (
    SlicesProjectClient as SlicesCoreProjectClient,
)
from slices_clientlib_core.util.pagination import PagedCollection

from .aiis_job_debug_logs import AIJobDebugLogsClient
from .aiis_job_output import AIJobOutputClient
from .schema.job import Job, JobStatus
from .schema.new_job import NewJob, NewJobReq
from .util.msgspec import json_encoder

_job_decoder = JsonDecoder(Job)


class AIJobClient:
    """AI Infrastructure Service Job API Client."""

    API_BASE_PATH = "/apis/ai.slices-ri.eu/v1/"

    def __init__(
        self,
        site_id: str,
        api_entry: ApiEntry,
        client: Client,
        slices_client: SlicesCoreProjectClient | SlicesCoreExperimentClient,
        slices_root_client: SlicesCoreClient,
    ) -> None:
        """Create AI Infrastructure Service Jobs Client."""
        self._site_id = site_id
        self._api_entry = api_entry
        self._client = client
        self._slices_client = slices_client
        self._slices_root_client = slices_root_client

        self._base_path = f"{self.API_BASE_PATH}jobs/"

        assert client.base_url is not None, "Expected base_url to be set"

    @cached_property
    def output(self) -> AIJobOutputClient:
        """Job Output Client."""
        return AIJobOutputClient(self._client)

    @cached_property
    def logs(self) -> AIJobDebugLogsClient:
        """Job Debug Logs Client."""
        return AIJobDebugLogsClient(self._client)

    def get(self, job: Job | str | UUID) -> Job:
        """Get job by ID."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        response = self._client.get(f"{self._base_path}{job_id}")

        try:
            response.raise_for_status()
            return _job_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def update_min_duration(self, job: Job | str | UUID, new_min_duration: str | timedelta):
        """Update job minDuration. Note that not all changes will be allowed."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        dur = (
            f"{new_min_duration.total_seconds()} s"
            if isinstance(new_min_duration, timedelta)
            else str(new_min_duration)
        )

        response = self._client.put(
            f"{self._base_path}{job_id}/request/scheduling/minDuration",
            headers={"Content-Type": "text/plain"},
            content=dur.encode("utf-8"),
        )

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def update_max_duration(self, job: Job | str | UUID, new_max_duration: str | timedelta):
        """Update job maxDuration. Note that not all changes will be allowed."""

        if not isinstance(self._slices_client, SlicesCoreExperimentClient):
            raise RuntimeError("Can only change Job maxDuration within an experiment context.")  # noqa: TRY004

        job_id = job.uuid if isinstance(job, Job) else str(job)
        dur = (
            f"{new_max_duration.total_seconds()} s"
            if isinstance(new_max_duration, timedelta)
            else str(new_max_duration)
        )

        response = self._client.put(
            f"{self._base_path}{job_id}/request/scheduling/maxDuration",
            headers={"Content-Type": "text/plain"},
            content=dur.encode("utf-8"),
        )

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def update_not_after(self, job: Job | str | UUID, new_not_after: datetime):
        """Update job notAfter."""

        if not isinstance(self._slices_client, SlicesCoreExperimentClient):
            raise RuntimeError("Can only change Job notAfter within an experiment context.")  # noqa: TRY004

        job_id = job.uuid if isinstance(job, Job) else str(job)

        response = self._client.put(
            f"{self._base_path}{job_id}/request/scheduling/notAfter",
            headers={"Content-Type": "text/plain"},
            content=new_not_after.isoformat(sep="T", timespec="seconds").encode("utf-8"),
        )

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def _update_status(self, job: Job | str | UUID, new_status: JobStatus):
        """Update job status. Only certain values allowed."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        response = self._client.put(
            f"{self._base_path}{job_id}/state/status",
            headers={"Content-Type": "text/plain"},
            content=str(new_status.name).encode("utf-8"),
        )

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def cancel(self, job: Job | str | UUID):
        """Cancel a job."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        self._update_status(job_id, JobStatus.CANCELLED)

    def halt(self, job: Job | str | UUID):
        """Manually HALT a job."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        self._update_status(job_id, JobStatus.MUSTHALT)

    def hold(self, job: Job | str | UUID):
        """Hold a QUEUED job so it cannot run."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        self._update_status(job_id, JobStatus.ONHOLD)

    def requeue(self, job: Job | str | UUID):
        """Requeue a job that is ONHOLD."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        self._update_status(job_id, JobStatus.QUEUED)

    def delete(self, job: Job | str | UUID) -> Job:
        """Delete compute job."""
        job_id = job.uuid if isinstance(job, Job) else str(job)
        response = self._client.delete(f"{self._base_path}{job_id}")

        try:
            response.raise_for_status()
            return _job_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def list(  # noqa: PLR0912, PLR0913, C901 we support all search options, and there are a lot.
        self,
        *,
        cluster_id: None | list[int] | int = None,
        allowed_states: None | list[JobStatus | str] | JobStatus | str = None,
        user_urn: None | list[str] | str = None,
        user_id: None | list[str] | str = None,
        user_name: None | list[str] | str = None,
        project_urn: None | list[str] | str = None,
        project_id: None | list[str] | str = None,
        project_name: None | list[str] | str = None,
        assigned_slave_name: None | list[str] | str = None,
        assigned_slave_instance_id: None | list[str] | str = None,
        interactive: bool | None = None,
        waste: bool | None = None,
        reservation_id: None | str | UUID = None,
    ) -> Collection[Job]:
        """List jobs.

        Be careful, this returns ALL jobs, there's no limit!
        They are returned as a "PagedCollection", which means they will only fetch jobs as required.
        You can use len() to get the total number of jobs, without requesting all of them.
        """

        def _list_or_single(v: list[Any] | Any) -> str:
            if isinstance(v, list):
                return ",".join([str(i) for i in v])
            return str(v)

        def _list_or_single_enum(v: list[Enum] | Enum) -> str:
            if isinstance(v, list):
                return ",".join([i.name for i in v])
            return v.name

        def _list_or_single_jobstatus_or_str(v: list[JobStatus | str] | JobStatus | str) -> str:
            if isinstance(v, str):
                return JobStatus(v).name
            if isinstance(v, JobStatus):
                return v.name
            if isinstance(v, list):
                return ",".join([i.name if isinstance(i, JobStatus) else JobStatus(str(i)).name for i in v])
            raise TypeError(f"unsupported type {type(v)}")

        params = {}
        if cluster_id is not None:
            params["cluster_id"] = _list_or_single(cluster_id)
        if allowed_states is not None:
            params["allowed_states"] = _list_or_single_jobstatus_or_str(allowed_states)
        if user_urn is not None:
            params["user_urn"] = _list_or_single(user_urn)
        if user_id is not None:
            params["user_id"] = _list_or_single(user_id)
        if user_name is not None:
            params["user_name"] = _list_or_single(user_name)
        if project_urn is not None:
            params["project_urn"] = _list_or_single(project_urn)
        if project_id is not None:
            params["project_id"] = _list_or_single(project_id)
        if project_name is not None:
            params["project_name"] = _list_or_single(project_name)
        if assigned_slave_name is not None:
            params["assigned_slave_name"] = _list_or_single(assigned_slave_name)
        if assigned_slave_instance_id is not None:
            params["assigned_slave_instance_id"] = _list_or_single(assigned_slave_instance_id)
        if interactive is not None:
            params["interactive"] = "true" if interactive else "false"
        if waste is not None:
            params["waste"] = "true" if waste else "false"
        if reservation_id is not None:
            params["reservation_id"] = str(reservation_id)

        return PagedCollection("job", Job, self._base_path, self._client, params)

    def create(self, new_job: NewJob) -> Job:
        """Create new jobs."""

        if not isinstance(self._slices_client, SlicesCoreExperimentClient):
            raise RuntimeError("Can only create AI Jobs within an experiment context.")  # noqa: TRY004

        id_token = self._slices_root_client.get_id_token(self._api_entry.audience, self._slices_client.project_id)

        # note: self._client has authentication with experiment token setup
        response = self._client.post(
            self._base_path,
            headers={"Content-Type": "application/json"},
            content=json_encoder.encode(NewJobReq(job=new_job, idtoken=id_token)),
        )

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

        return _job_decoder.decode(response.content)



"""AI Infrastructure Service API Client."""

from functools import cached_property

from httpx import Client
from slices_clientlib_core.base.config import ApiEntry
from slices_clientlib_core.client import (
    SlicesClient as SlicesCoreClient,
)
from slices_clientlib_core.client import (
    SlicesExperimentClient as SlicesCoreExperimentClient,
)
from slices_clientlib_core.client import (
    SlicesProjectClient as SlicesCoreProjectClient,
)

from slices_clientlib_aiis.aiis_job import AIJobClient
from slices_clientlib_aiis.aiis_news import AINewsClient
from slices_clientlib_aiis.aiis_workers import AIWorkersClient


class AIInfraServiceClient:
    """AI Infrastructure Service Client."""

    def __init__(
        self,
        site_id: str,
        api_entry: ApiEntry,
        client: Client,
        slices_client: SlicesCoreProjectClient | SlicesCoreExperimentClient,
        slices_root_client: SlicesCoreClient,
    ) -> None:
        """Create AI Infrastructure Service Client."""
        self._site_id = site_id
        self._api_entry = api_entry
        self._client = client
        self._slices_client = slices_client
        self._slices_root_client = slices_root_client
        assert client.base_url is not None, "Expected base_url to be set"

    @cached_property
    def jobs(self) -> AIJobClient:
        """AI Jobs Client."""
        return AIJobClient(self._site_id, self._api_entry, self._client, self._slices_client, self._slices_root_client)

    @cached_property
    def news(self) -> AINewsClient:
        """AIIS News Client."""
        return AINewsClient(self._client)

    @cached_property
    def workers(self) -> AIWorkersClient:
        """AIIS Worker Client."""
        return AIWorkersClient(self._client)



"""Slices AI Infrastructure Service Resource API."""

from httpx import Client, HTTPError
from msgspec.json import Decoder as JsonDecoder
from slices_clientlib_core.base.exceptions import from_httpx_exception

from .schema.news import News

_news_decoder = JsonDecoder(News)
_news_list_decoder = JsonDecoder(list[News])


class AINewsClient:
    """AI Infrastructure Service News API Client."""

    API_BASE_PATH = "/apis/ai.slices-ri.eu/v1/news/"

    def __init__(self, client: Client) -> None:
        """Create AI Infrastructure Service News Client."""

        self._client = client

        self._base_path = f"{self.API_BASE_PATH}"

        assert client.base_url is not None, "Expected base_url to be set"

    def get(self, id: str) -> News:
        """Get news by ID."""
        response = self._client.get(f"{self._base_path}{id}")

        try:
            response.raise_for_status()
            return _news_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def list(self) -> list[News]:
        """List news."""

        response = self._client.get(f"{self._base_path}")

        try:
            response.raise_for_status()
            return _news_list_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err



"""Constants."""

CONFIG_DIR = ".slices"

SLICES_AI_CONFIG_CACHE_FILENAME = "cache_slices-ai-config.json"
DEFAULT_SLICES_AI_CONFIG_URL = "https://portal.slices-ri.eu/ai-service-config.json"
SLICES_AI_CONFIG_CACHE_TIMEOUT_MINUTES = 5



"""Utility class to setup Fabric connections to compute resources.

More information about fabric: https://docs.fabfile.org .
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import contextlib

    with contextlib.suppress(ImportError):
        # fabric is an optional dependency, so we only do an optional import for type checking here
        from fabric import Connection  # pyright: ignore [reportMissingImports]  # noqa: PLC0415

    from slices_clientlib_aiis.schema.job import SSHInfo


def connect(ssh_info: "SSHInfo") -> "Connection":
    """Create a Fabric connection."""
    try:
        from fabric import Connection  # pyright: ignore [reportMissingImports]  # noqa: PLC0415
    except ImportError as err:
        raise ImportError("To use this function, you need to install the 'fabric' package") from err

    return Connection(
        host=ssh_info.host,
        port=ssh_info.port,
        user=ssh_info.username,
        gateway=connect(ssh_info.jump_proxy) if ssh_info.jump_proxy else None,
    )



"""Slices Client."""

import importlib.metadata
from pathlib import Path

import httpx
import msgspec
from httpx import HTTPError
from slices_clientlib_core.base.config import SlicesCentralApiClientsConfig
from slices_clientlib_core.base.exceptions import from_httpx_exception
from slices_clientlib_core.client import (
    SlicesClient as SlicesCoreClient,
)
from slices_clientlib_core.client import (
    SlicesExperimentClient as SlicesCoreExperimentClient,
)
from slices_clientlib_core.client import (
    SlicesProjectClient as SlicesCoreProjectClient,
)

from .aiis import AIInfraServiceClient
from .base.config import SlicesAIApiClientsConfig, load_default_aiis_api_clients_config


def _load_file_or_url(location: str, client: httpx.Client) -> str:
    if location.startswith("http"):
        response = client.get(location, headers={"Accept": "application/json"})
        try:
            response.raise_for_status()
            return response.content.decode()
        except HTTPError as err:
            raise from_httpx_exception(err) from err
    elif Path(location).exists():
        with Path(location).open() as f:
            return f.read()
    else:
        raise ValueError(f"{location!r} is not a valid path or URL")


def _get_http_user_agent():
    """Get HTTP User Agent."""
    return f"slices-aiis-client/{importlib.metadata.version('slices_clientlib_aiis')}"


DEFAULT_HTTP_USER_AGENT = _get_http_user_agent()


def _get_site_id(site_or_infra_id: str) -> str:
    """Get site ID from site or infrastructure ID."""

    id_parts = site_or_infra_id.split("-")

    if len(id_parts) > 2 and id_parts[2] == "ai":  # noqa: PLR2004
        # It's an infrastructure/cluster ID: {site_id}-ai-{cluster_id}
        return f"{id_parts[0]}-{id_parts[1]}"
    # Assume it's a site ID
    return site_or_infra_id


class SlicesClientWithAIInfraService(SlicesCoreClient):
    """Slices Client, with support for AI Infrastructure Service."""

    _aiis_api_clients_config: SlicesAIApiClientsConfig

    def __init__(
        self,
        central_config: SlicesCentralApiClientsConfig | str | None = None,
        ai_config: SlicesAIApiClientsConfig | str | None = None,
        ignore_config_cache: bool = False,
        http_user_agent: str = DEFAULT_HTTP_USER_AGENT,
    ) -> None:
        """Create a Slices AIIS client."""

        super().__init__(
            central_config=central_config,
            ignore_config_cache=ignore_config_cache,
            http_user_agent=http_user_agent,
        )

        self._init_param_ai_config = ai_config

        # Load AIIS config
        if ai_config and isinstance(ai_config, str):
            self._aiis_api_clients_config = msgspec.json.decode(
                _load_file_or_url(ai_config, self._no_auth_client), type=SlicesAIApiClientsConfig
            )
        elif ai_config and isinstance(ai_config, SlicesAIApiClientsConfig):
            self._aiis_api_clients_config = ai_config
        elif ai_config is None:
            self._aiis_api_clients_config = load_default_aiis_api_clients_config(self._no_auth_client)
        else:
            raise TypeError(f"Unsupported ai_config {ai_config!r}")

    def reload_ai_config(self) -> bool:
        """Reload the AIIS slices config from the default online source.

        This is useful because it is locally cached.

        This does nothing if the central slices config was not loaded from
        the default online source or its cached version.

        Returns: True if reload was done, False if reload was skipped.
        """
        if self._init_param_ai_config is not None:
            return False
        self._aiis_api_clients_config = load_default_aiis_api_clients_config(self._no_auth_client, ignore_cache=True)
        return True

    @property
    def ai_api_clients_config(self) -> SlicesAIApiClientsConfig:
        """Get AIIS API Clients Config."""
        return self._aiis_api_clients_config

    def for_project(self, project_id: str) -> "SlicesProjectClientWithAIInfraService":
        """Get Slices Project Client."""
        return SlicesProjectClientWithAIInfraService(project_id, self)

    def for_experiment(self, experiment_id: str, project_id: str) -> "SlicesExperimentClientWithAIInfraService":
        """Get Slices Project Client."""
        return SlicesExperimentClientWithAIInfraService(project_id, experiment_id, self)


class SlicesProjectClientWithAIInfraService(SlicesCoreProjectClient):
    """Slices client for a specific project, with support for AI Infrastructure Service."""

    def __init__(self, project_id: str, generic_client: SlicesClientWithAIInfraService) -> None:
        """Create Slices Project Client, with support for AI Infrastructure Service."""
        super().__init__(project_id, generic_client)
        if not project_id.startswith("proj_"):
            raise ValueError(f"Invalid project ID: {project_id!r}")
        self._aiis_api_clients_config = generic_client._aiis_api_clients_config
        self._generic_aiis_client = generic_client

    def for_experiment(self, experiment_id: str) -> "SlicesExperimentClientWithAIInfraService":
        """Get Slices Project Client."""
        return SlicesExperimentClientWithAIInfraService(experiment_id, self.project_id, self._generic_aiis_client)

    def ai(self, site_or_infra_id: str | None = None) -> AIInfraServiceClient:
        """Get AI Infrastructure Service API Client for a specific site."""

        if site_or_infra_id:
            site_id = _get_site_id(site_or_infra_id)
        else:
            site_id = self._aiis_api_clients_config.ai_service_default_site

        aiis_api_entry = self._aiis_api_clients_config.ai_service_sites.get(site_id)

        if not aiis_api_entry:
            raise ValueError(f"Unknown Site ID: {site_id}")

        return AIInfraServiceClient(
            site_id,
            aiis_api_entry,
            self._generic_client._create_httpx_client_for_api_entry(
                api_entry=aiis_api_entry, project_id=self.project_id
            ),
            self,
            self._generic_aiis_client,
        )


class SlicesExperimentClientWithAIInfraService(SlicesCoreExperimentClient):
    """Slices client for a specific experiment, with support for AI Infrastructure Service."""

    def __init__(self, experiment_id: str, project_id: str, generic_client: SlicesClientWithAIInfraService) -> None:
        """Create Slices Experiment Client, with support for AI Infrastructure Service."""
        super().__init__(experiment_id, project_id, generic_client)
        if not experiment_id.startswith("exp_"):
            raise ValueError(f"Invalid experiment ID: {experiment_id!r}")
        if not project_id.startswith("proj_"):
            raise ValueError(f"Invalid project ID: {project_id!r}")

        self._aiis_api_clients_config = generic_client._aiis_api_clients_config
        self._generic_aiis_client = generic_client

    def ai(self, site_or_infra_id: str | None = None) -> AIInfraServiceClient:
        """Get AI Infrastructure Service API Client for a specific site."""

        if site_or_infra_id:
            site_id = _get_site_id(site_or_infra_id)
        else:
            site_id = self._aiis_api_clients_config.ai_service_default_site

        ai_api_entry = self._aiis_api_clients_config.ai_service_sites.get(site_id)

        if not self.project_id:
            raise ValueError("Problem with Project ID")
        if not self.experiment_id:
            raise ValueError("Problem with Experiment ID")

        if not ai_api_entry:
            raise ValueError(f"Unknown site ID: {site_id}")

        client = self._generic_client._create_httpx_client_for_api_entry(
            api_entry=ai_api_entry, project_id=self.project_id, experiment_id=self.experiment_id
        )
        return AIInfraServiceClient(site_id, ai_api_entry, client, self, self._generic_aiis_client)



"""Schema for News items."""

from typing import Literal

from msgspec import Struct

from slices_clientlib_aiis.util.msgspec import AwareDatetime


class News(Struct, rename="camel"):
    """News Item."""

    id: str
    created: AwareDatetime
    enabled: bool
    type: Literal["WARNING", "INFO", "CRITICAL"]
    title: str
    text: str
    tags: list[str]
    not_before: AwareDatetime | None
    not_after: AwareDatetime | None



"""Schema for Worker/Slave and Cluster info."""

from msgspec import Struct, field

from slices_clientlib_aiis.util.msgspec import AwareDatetime


class GpuModel(Struct, rename="camel"):
    """GPU Model."""

    vendor: str
    name: str
    memory_mb: int


class ResourceInfo(Struct, rename="camel"):
    """Resource Information."""

    system_total: int
    acquired: int
    used: int
    available: int

    @classmethod
    def zero(cls):
        """Return instance with everything set to 0."""
        return cls(
            system_total=0,
            acquired=0,
            used=0,
            available=0,
        )


class ActiveJobsCounts(Struct, rename="camel"):
    """Return the number of jobs in any of the 'active' job states.

    Can be used for both clusters and slaves.
    """

    onhold: int
    queued: int
    assigned: int
    starting: int
    running: int
    musthalt: int
    halting: int

    @classmethod
    def zero(cls):
        """Return an ActiveJobsCounts with all counters on zero."""
        return cls(
            onhold=0,
            queued=0,
            assigned=0,
            starting=0,
            running=0,
            musthalt=0,
            halting=0,
        )


class Slave(Struct, rename="camel"):
    """Slave/Worker Info."""

    name: str
    instance_id: str
    cluster_id: int
    pid: int | None
    software_version: str | None

    aliases: list[str]
    comment: str | None
    host: str | None

    gpu_model: list[GpuModel]
    cpu_model: list[str]
    worker: ResourceInfo
    cpu_memory_mb: ResourceInfo
    gpu: ResourceInfo
    cpu: ResourceInfo
    last_update: AwareDatetime
    shutting_down: bool
    cuda_version_full: str | None = None
    cuda_version_major: int | None = None

    docker_disk_used_percent: float = -1
    accepting_jobs: bool = True
    active_jobs_counts: ActiveJobsCounts | None = None

    storage_paths_available: list[str] | None = None
    storage_aliases_available: dict[str, str] | None = None
    storage_paths_available_project_limited: dict[str, list[str]] | None = None
    active_job_uuids: list[str] | None = None


class Cluster(Struct, rename="camel"):
    """Cluster Info."""

    cluster_id: int
    site_id: str
    comment: str
    is_private: bool
    slave_count: int = 0
    gpu_model: list[GpuModel] = []
    worker: ResourceInfo = field(default_factory=ResourceInfo.zero)
    cpu_memory_mb: ResourceInfo = field(default_factory=ResourceInfo.zero)
    gpu: ResourceInfo = field(default_factory=ResourceInfo.zero)
    cpu: ResourceInfo = field(default_factory=ResourceInfo.zero)
    active_jobs_counts: ActiveJobsCounts | None = None
    have_access: bool | None = None
    allowed_projects: list[str] | None = None
    storage_paths_available: list[str] | None = None
    storage_aliases_available: dict | None = None



"""Schema for Job."""

from datetime import datetime, timedelta
from enum import Enum
from typing import Annotated, Optional
from uuid import UUID

import pytz
from msgspec import Meta, Struct, field

from slices_clientlib_aiis.schema.job_statistics import GPULabUsageStatistics, WasteReview
from slices_clientlib_aiis.util.fabric import connect as connect_with_fabric
from slices_clientlib_aiis.util.msgspec import AwareDatetime, EmailStr, NonEmptyStr, PositiveInt, PublicSshKey

DurationStr = Annotated[str, Meta(min_length=2)]


class JobStatus(Enum):
    """Job state."""

    ONHOLD = "On Hold"
    QUEUED = "Queued"
    ASSIGNED = "Assigned"
    STARTING = "Starting"
    RUNNING = "Running"
    CANCELLING = "Cancelling"
    CANCELLED = "Cancelled"
    FINISHED = "Finished"
    DELETED = "Deleted"
    FAILED = "Failed"
    MUSTHALT = "MustHalt"
    HALTING = "Halting"
    HALTED = "Halted"

    @classmethod
    def _missing_(cls, value):
        """Handle missing enum values by ignoring case."""
        if isinstance(value, cls):
            return value
        if not isinstance(value, str):
            return None
        for member in cls:
            if member.name.lower() == value.lower():
                return member
            if member.value.lower() == value.lower():
                return member
        return None


class JobStorage(Struct, rename="camel"):
    """Job Storage (tmpfs or dictionary)."""

    container_path: NonEmptyStr | None = None
    host_path: NonEmptyStr | None = None
    size_gb: PositiveInt | None = None  # only for tmpfs


class JobPortMapping(Struct, rename="camel"):
    """Job Port Mapping."""

    container_port: int
    host_port: int | None = None
    host_ip: str | None = None


class JobRequestDocker(Struct, rename="camel"):
    """Job Request: Docker."""

    image: NonEmptyStr
    command: str | list[str] = []
    environment: dict[NonEmptyStr, str | int | bool] = {}
    storage: list[JobStorage] = []
    port_mappings: list[JobPortMapping] = []
    project_gid_variable_name: NonEmptyStr | None = None
    working_dir: NonEmptyStr | None = None
    group_add: list[NonEmptyStr] = []
    user: NonEmptyStr | None = None


class MaxSimultaneousJobs(Struct, rename="camel"):
    """Maximum number of simultaneous jobs."""

    bucket_name: str
    bucket_max: int


class JobRequestScheduling(Struct, rename="camel"):
    """GPULab scheduling instructions Base."""

    min_duration: DurationStr
    max_duration: DurationStr
    interactive: bool = False
    restartable: bool = False
    allow_halt_without_signal: bool = False
    reservation_ids: list[UUID] = []
    max_simultaneous_jobs: MaxSimultaneousJobs | None = None
    not_before: AwareDatetime | None = None
    not_after: AwareDatetime | None = None


class JobRequestResources(Struct, rename="camel"):
    """Job Request Resources."""

    cpus: PositiveInt
    gpus: int
    cpu_memory_gb: PositiveInt
    gpu_memory_gb: PositiveInt | None = None

    features: list[NonEmptyStr] = []
    gpu_model: list[NonEmptyStr] = []

    min_cuda_version: PositiveInt | None = None
    cluster_id: PositiveInt | None = None
    slave_name: NonEmptyStr | None = None
    slave_instance_id: NonEmptyStr | None = None


class JobRequestExtra(Struct, rename="camel"):
    """Job Request Extras."""

    ssh_pub_keys: list[PublicSshKey] = []
    email_on_queue: list[EmailStr] = []
    email_on_run: list[EmailStr] = []
    email_on_end: list[EmailStr] = []
    email_on_halt: list[EmailStr] = []
    email_on_restart: list[EmailStr] = []


class JobRequestBase(Struct, rename="camel"):
    """Job Request Base."""

    resources: JobRequestResources
    docker: JobRequestDocker


class JobRequest(JobRequestBase):
    """Job Request Base."""

    scheduling: JobRequestScheduling
    extra: JobRequestExtra = field(default_factory=lambda: JobRequestExtra())


class JobOwner(Struct, rename="camel"):
    """Job Owner."""

    user_id: NonEmptyStr
    project_id: NonEmptyStr
    experiment_id: NonEmptyStr | None


class RestartInfo(Struct, rename="camel"):
    """Restart Information.

    Available in jobs that have been restarted after being halted.
    """

    initial_job_uuid: UUID
    restart_count: int


class HaltReason(Enum):
    """Halt Reason."""

    UNKNOWN = "Unknown"
    USER = "User requested fallback manually"
    ADMIN = "Admin requested fallback"
    SCORE = "Score"
    RESERVATION = "Reservation Priority"
    CLUSTER = "Private Cluster Priority"

    @classmethod
    def _missing_(cls, value):
        """Handle missing enum values by ignoring case."""
        if isinstance(value, cls):
            return value
        if not isinstance(value, str):
            return None
        for member in cls:
            if member.name.lower() == value.lower():
                return member
            if member.value.lower() == value.lower():
                return member
        return None


class HaltInfo(Struct, rename="camel"):
    """Halt Info."""

    reason: HaltReason
    time: datetime
    started_job_uuids: list[str]
    halted_job_uuids: list[str]
    slave_name: str
    slave_instance: str
    started_job_scores: list[float]
    halted_job_scores: list[float]


class JobStateScheduling(Struct, rename="camel"):
    """Job State Scheduling."""

    assigned_cluster_id: int | None = None
    assigned_instance_id: str | None = None
    assigned_slave_name: str | None = None
    queued_explanations: list[str] = []
    tally_increment: float | None = None
    halt_events: list[HaltInfo] = []
    ignoring_reservation_ids: list[str] = []


class JobEventTimes(Struct, rename="camel"):
    """Job Event Times."""

    created: AwareDatetime
    status_updated: AwareDatetime
    QUEUED: AwareDatetime | None = field(name="QUEUED", default=None)
    ASSIGNED: AwareDatetime | None = field(name="ASSIGNED", default=None)
    STARTING: AwareDatetime | None = field(name="STARTING", default=None)
    RUNNING: AwareDatetime | None = field(name="RUNNING", default=None)
    FINISHED: AwareDatetime | None = field(name="FINISHED", default=None)
    FAILED: AwareDatetime | None = field(name="FAILED", default=None)
    CANCELLED: AwareDatetime | None = field(name="CANCELLED", default=None)
    DELETED: AwareDatetime | None = field(name="DELETED", default=None)
    MUSTHALT: AwareDatetime | None = field(name="MUSTHALT", default=None)
    HALTING: AwareDatetime | None = field(name="HALTING", default=None)
    HALTED: AwareDatetime | None = field(name="HALTED", default=None)
    long_run_notify: AwareDatetime | None = None
    waste_notify: AwareDatetime | None = None

    @property
    def end_date(self) -> AwareDatetime | None:
        """End date."""
        if self.FINISHED:
            return self.FINISHED
        if self.HALTED:
            return self.HALTED
        if self.FAILED:
            return self.FAILED
        if self.CANCELLED:
            return self.CANCELLED
        if self.DELETED:
            return self.DELETED
        return None

    @property
    def duration_or_none(self) -> timedelta | None:
        """Get current duration of job, or None if not started."""
        start = self.RUNNING
        if start is None:
            return None

        end = self.FINISHED or self.CANCELLED or self.FAILED or self.HALTED or self.DELETED
        if end is None:
            end = datetime.now(pytz.utc)
        return end - start

    @property
    def duration_or_zero(self) -> timedelta:
        """Get current duration of job, or 0 duration if not started."""
        return self.duration_or_none or timedelta(seconds=0)

    @property
    def queue_duration(self) -> timedelta:
        """Get queue duration, or zero if not ever queued."""
        start = self.QUEUED
        if start is None:
            return timedelta(seconds=0)

        end = self.STARTING or self.RUNNING or self.FAILED or self.HALTED or self.CANCELLED or self.DELETED
        if end is None:
            end = datetime.now(pytz.utc)

        return end - start


class GpuInfo(Struct, rename="camel"):
    """GPU Information."""

    index: int
    uuid: str
    serial: str
    name: str
    brand: str
    minor_number: int
    board_id: int
    bridge_chip_info: str
    is_multi_gpu_board: bool
    max_pcie_link_generation: int
    max_pcie_link_width: int
    vbios_version: str


class GpuOverview(Struct, rename="camel"):
    """GPU Information Overview."""

    cuda_version_full: str
    cuda_version_int: int
    cuda_version_major: int
    cuda_version_minor: int | float
    driver_version: str
    nvml_version: str
    gpus: list[GpuInfo]


class JobStateResources(Struct, rename="camel"):
    """Job State Resources."""

    cluster_id: PositiveInt
    gpu_ids: list[int]
    cpu_ids: list[int]
    cpu_memory_gb: int
    """does NOT include tmpfs memory (it is listed separately below)"""
    gpu_memory_gb: int
    """if less than a GB, this is rounded DOWN (so it can be 0!)"""

    slave_name: str
    slave_host: str
    slave_instance_id: str
    slave_instance_pid: int
    worker_id: int

    site_id: str
    infra_id: str
    machine_id: str

    ssh_host: str | None = None
    ssh_port: int | None = None
    ssh_username: str | None = None
    ssh_proxy_host: str | None = None
    ssh_proxy_port: int | None = None
    ssh_proxy_username: str | None = None

    port_mappings: list[JobPortMapping] = []
    gpu_details: GpuOverview | None = None

    tmpfs_mem_gb: int = 0

    @property
    def cpu_memory_mb(self) -> int:
        """CPU Memory in MB."""
        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc.), for memory, base 1024 is always used.
        # (and docker seems to use 1000 based, which is silly)
        return self.cpu_memory_gb * 1024

    @property
    def cpu_memory_byte(self) -> int:
        """CPU Memory in bytes."""
        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc.), for memory, base 1024 is always used.
        # (and docker seems to use 1000 based, which is silly)
        return self.cpu_memory_gb * 1024 * 1024 * 1024

    @property
    def tmpfs_mem_mb(self) -> int:
        """Tmpfs Memory in MB."""
        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc.), for memory, base 1024 is always used.
        # (and docker seems to use 1000 based, which is silly)
        return (self.tmpfs_mem_gb or 0) * 1024

    @property
    def tmpfs_mem_byte(self) -> int:
        """Tmpfs Memory in bytes."""
        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc.), for memory, base 1024 is always used.
        # (and docker seems to use 1000 based, which is silly)
        return (self.tmpfs_mem_gb or 0) * 1024 * 1024 * 1024

    @property
    def all_cpu_memory_gb(self) -> int:
        """All Memory in GB."""
        return self.cpu_memory_gb + (self.tmpfs_mem_gb or 0)

    @property
    def all_cpu_memory_mb(self) -> int:
        """All Memory in MB."""
        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc.), for memory, base 1024 is always used.
        # (and docker seems to use 1000 based, which is silly)
        return (self.all_cpu_memory_gb or 0) * 1024

    @property
    def all_cpu_memory_byte(self) -> int:
        """All Memory in bytes."""
        # while GB is sometimes 1000^3 instead of 1024^3 (typically HD size etc.), for memory, base 1024 is always used.
        # (and docker seems to use 1000 based, which is silly)
        return (self.all_cpu_memory_gb or 0) * 1024 * 1024 * 1024


class JobState(Struct, rename="camel"):
    """Current Job State.

    State vs Status
       - In the english language: very related, and often synonyms
       - Typically in technical language, state is used in a more broad sense, and status is more "one-dimensional"
           -> thus you can have a state containing multiple statuses

      Here:
       - In JobState we describe the entire variable state of the Job.
         "status" holds the ID of the current discrete step in the Job's lifecycle ("the workflow of job execution")
         so "lifecycle_step" or "workflow_position" would be a synonym for our "status", but both feels too convoluted
    """

    status: JobStatus
    scheduling: JobStateScheduling
    event_times: JobEventTimes
    resources: JobStateResources | None = None
    final_usage_statistics: GPULabUsageStatistics | None = None
    waste_review: WasteReview | None = None


class SSHInfo(Struct):
    """Info about SSH Login to a resource."""

    host: str
    port: int
    username: str

    host_keys: list[str] | None
    """The public key(s) of the target SSH server, in openssh format.

    This list can be empty. It is recommended that it is not.
    The typical line you need to add to ~/.ssh/known_hosts is the host, a space, and then one of these keys.
    """

    description: str | None = None
    jump_proxy: Optional["SSHInfo"] = None

    def connect(self):
        """Create a Fabric connection."""

        return connect_with_fabric(self)


class Job(Struct, rename="camel"):
    """AI Infrastructure Service Job."""

    uuid: UUID
    slices_id: str
    name: str
    request: JobRequest
    owner: JobOwner
    state: JobState
    description: str | None = None
    restart_info: RestartInfo | None = None

    @property
    def ssh_info(self) -> list[SSHInfo]:
        """Return a list of SSHInfo objects for this job.

        These objects are compatible with the slices BI SSHInfo objects.
        """

        if (
            self.state.resources is None
            or self.state.resources.ssh_host is None
            or self.state.resources.ssh_username is None
        ):
            return []
        proxy: SSHInfo | None = None
        if self.state.resources.ssh_proxy_host is not None and self.state.resources.ssh_proxy_username is not None:
            proxy = SSHInfo(
                host=self.state.resources.ssh_proxy_host,
                port=self.state.resources.ssh_proxy_port or 22,
                username=self.state.resources.ssh_proxy_username,
                host_keys=None,
                description=None,
                jump_proxy=None,
            )
        return [
            SSHInfo(
                host=self.state.resources.ssh_host,
                port=self.state.resources.ssh_port or 22,
                username=self.state.resources.ssh_username,
                host_keys=None,
                description=None,
                jump_proxy=proxy,
            )
        ]

    @property
    def id(self) -> UUID:
        """Job ID alias."""
        return self.uuid

    @property
    def short_id(self) -> str | None:
        """Short ID that is derived from full uuid."""
        if not self.id:
            return None
        id_s = str(self.id)
        if "-" in id_s:
            return id_s[: id_s.index("-")]
        return id_s



"""Utils for httpx lib."""

import json
from contextlib import contextmanager

from httpx import HTTPStatusError


@contextmanager
def print_httpx_exception_body():
    """When httpx exceptions are raised, print the response body."""

    try:
        yield
    except HTTPStatusError as err:
        if err.response.headers.get("Content-Type") == "application/json":
            print(f"HTTP error json: {json.dumps(err.response.json(), indent=2)}")  # noqa: T201
        else:
            print(f"HTTP error json: {err.response.text}")  # noqa: T201
        raise



"""Job Debug Logs."""

from datetime import datetime
from enum import Enum
from typing import Any
from uuid import UUID

import pytz
from msgspec import Struct, field

from slices_clientlib_aiis.schema.job import JobStatus
from slices_clientlib_aiis.util.msgspec import AwareDatetime


class JobDebugLogType(Enum):
    """Job Event Type."""

    STATUS_CHANGE = "Status Change"
    DEBUG = "Debug"
    INFO = "Info"
    WARN = "Warning"
    ERROR = "Error"

    @classmethod
    def _missing_(cls, value: Any):
        """Handle missing enum values by ignoring case."""
        if isinstance(value, cls):
            return value
        if not isinstance(value, str):
            return None
        for member in cls:
            if member.name.lower() == value.lower():
                return member
            if member.value.lower() == value.lower():
                return member
        return None


class JobDebugLog(Struct, rename="camel"):
    """Job Debug Log aka Job Event."""

    job_id: UUID
    type: JobDebugLogType
    time: AwareDatetime | None = field(default_factory=lambda: datetime.now(tz=pytz.utc))
    new_state: JobStatus | None = None
    msg: str | None = None



"""Slices AIIS Client API."""

import logging

from .aiis import AIInfraServiceClient
from .client import (
    SlicesClientWithAIInfraService,
    SlicesExperimentClientWithAIInfraService,
    SlicesProjectClientWithAIInfraService,
)

__all__ = [
    "SlicesClientWithAIInfraService",
    "SlicesProjectClientWithAIInfraService",
    "SlicesExperimentClientWithAIInfraService",
    "AIInfraServiceClient",
]

# All code in this library uses standard python logging.
# But since this library root logger does not propagate, nothing is done with this logging by default.
# Any app using this library that want to access the logging needs to add an event handler to this logger.
# Example:
#   slices_clientlib_core_logger = logger.getLogger("slices_clientlib_core")
#   slices_clientlib_core_logger.setLevel(logging.DEBUG)
#   slices_clientlib_core_logger.addHandler(logging.StreamHandler())

logger = logging.getLogger(__name__)
NULL_HANDLER = logging.NullHandler()
logger.addHandler(NULL_HANDLER)  # Use NULL_HANDLER to prevent warning "No handlers could be found for logger XXX"
logger.propagate = False  # Do not propagate to parent loggers
logger.setLevel(logging.DEBUG)



"""Job Statistics."""

from datetime import datetime
from enum import Enum

from msgspec import Struct, field

from slices_clientlib_aiis.util.msgspec import AwareDatetime


class ContainerUsageStatistics(Struct, rename="camel"):
    """Container Usage Statistics."""

    first_time: datetime
    last_time: datetime
    agg_period_ns: int
    cpu_count: int
    cpu_usage: float  # in nr CPU's, so between 0 and cpu_count
    cpu_usage_total_ns: int
    cpu_usage_kernelmode_ns: int
    cpu_usage_usermode_ns: int
    max_pid_count: int
    mem_limit_byte: int
    mem_usage_byte: int
    mem_max_usage_byte: int
    network_rx_byte: int
    network_tx_byte: int
    host_total_mWh_used: int | None = field(name="hostTotalmWhUsed", default=None)
    host_average_mW_used: int | None = field(name="hostAveragemWUsed", default=None)
    cpu_total_mWh_used: int | None = field(name="cpuTotalmWhUsed", default=None)
    cpu_average_mW_used: int | None = field(name="cpuAveragemWUsed", default=None)

    total_mWh_used: int | None = field(name="totalmWhUsed", default=None)
    average_mW_used: float | None = field(name="averagemWUsed", default=None)


class GpuUsageStatistics(Struct, rename="camel"):
    """GPU Usage Statistics."""

    gpu_count: int
    average_utilization: float
    average_mem_utilization: float
    total_mWh_used: int | None = field(name="totalmWhUsed", default=None)
    average_mW_used: int | None = field(name="averagemWUsed", default=None)


class WasteVerdict(Enum):
    """Waste Verdict."""

    GOOD = "Good"
    """Job is not wasting resources (yet, for running jobs)"""
    UNCERTAIN = "Uncertain"
    """Job might be wasting resources, or might not"""
    WASTE = "Waste"
    """Job is wasting resources"""
    SHORT = "Short"
    """Job was too short to verdict. It's OK if short test/failed jobs waste resources."""
    UNDECIDED = "Undecided"
    """Too soon to tell if job wastes resources"""

    @classmethod
    def _missing_(cls, value):
        """Handle missing enum values by ignoring case."""
        if isinstance(value, cls):
            return value
        if not isinstance(value, str):
            return None
        for member in cls:
            if member.name.lower() == value.lower():
                return member
            if member.value.lower() == value.lower():
                return member
        return None


class WasteFlag(Enum):
    """Waste Flag."""

    RED = "Red Flag"
    """Certainly unacceptable resource waste"""
    ORANGE = "Orange Flag"
    """Quite likely unacceptable resource waste"""
    YELLOW = "Yellow Flag"
    """Resource waste, but probably not unacceptable"""

    @classmethod
    def _missing_(cls, value):
        """Handle missing enum values by ignoring case."""
        if isinstance(value, cls):
            return value
        if not isinstance(value, str):
            return None
        for member in cls:
            if member.name.lower() == value.lower():
                return member
            if member.value.lower() == value.lower():
                return member
        return None


class WasteFlagInfo(Struct, rename="camel"):
    """Waste Flag Info."""

    flag: WasteFlag
    explanation: str


class WasteReview(Struct, rename="camel"):
    """Waste Review."""

    updated: AwareDatetime
    final: bool
    """This is the final report, as the job has finished. It can include details intermediate reports don't."""
    verdict: WasteVerdict
    """Is this job wasting resources, or not?"""
    flags: list[WasteFlagInfo] = []
    """Why is this job wasting resources? (empty if not wasting)"""
    debug: dict[str, bool | int | float | str] = {}


class GPULabUsageStatistics(Struct, rename="camel"):
    """GPULab Usage Statistics."""

    container_statistics: ContainerUsageStatistics
    gpu_statistics: GpuUsageStatistics

    # Power measurements
    #    (None if not (yet) available/implemented)
    #    (and for backward compatibility default is 0.0)

    total_mWh_used: int | None = field(name="totalmWhUsed", default=None)
    average_mW_used: float | None = field(name="averagemWUsed", default=None)



"""CLI configuration management."""

import logging
from functools import cache

import httpx
from msgspec import Struct
from slices_clientlib_core.base.config import ApiEntry, cache_dir_location, load_config_from_url_or_cache

from slices_clientlib_aiis.base.const import (
    DEFAULT_SLICES_AI_CONFIG_URL,
    SLICES_AI_CONFIG_CACHE_FILENAME,
    SLICES_AI_CONFIG_CACHE_TIMEOUT_MINUTES,
)

logger = logging.getLogger(__name__)

slices_aiis_config_cache_file_location = cache_dir_location / SLICES_AI_CONFIG_CACHE_FILENAME


class SlicesAIApiClientsConfig(Struct):
    """Configuration model used to create slices AI api client objects."""

    ai_service_default_site: str
    """Default site ID to use when not specified by user."""

    ai_service_sites: dict[str, ApiEntry]
    """AI Service site info, by slices site ID."""


@cache
def load_default_aiis_api_clients_config(
    client: httpx.Client,
    *,
    ignore_cache: bool = False,
) -> SlicesAIApiClientsConfig:
    """Load default API clients configuration for the slices AI Service.

    This is loaded from the central URL, or from a locally cached copy.
    """
    return load_config_from_url_or_cache(
        slices_aiis_config_cache_file_location,
        DEFAULT_SLICES_AI_CONFIG_URL,
        SLICES_AI_CONFIG_CACHE_TIMEOUT_MINUTES,
        SlicesAIApiClientsConfig,
        client,
        ignore_cache,
    )
